class MyNeatoLoginException(Exception):
    def __init__(self, message):
        Exception.__init__(self)
        self.message = message

class MyNeatoRobotException(Exception):
    def __init__(self, message):
        Exception.__init__(self)
        self.message = message

class MyNeatoException(Exception):
    def __init__(self, message):
        Exception.__init__(self)
        self.message = message

class MyNeatoUnsupportedDevice(Exception):
    def __init__(self, message):
        Exception.__init__(self)
        self.message = message



import logging
import sys
import os

from pyneato import (
    Account,
    Neato,
    OrbitalPasswordSession,
)

Log_Format = "%(levelname)s %(asctime)s - %(message)s"

logging.basicConfig(
    stream = sys.stdout,
    filemode = "w",
    format = Log_Format,
    level = logging.DEBUG
)

vendor = Neato()

email = os.environ['MYNEATO_USER']
password = os.environ['MYNEATO_PASSWORD']

session = OrbitalPasswordSession(
    email,
    password
)
account = Account(session)
robot = account.robots.pop()

for index, floorplan in enumerate(account.floorplans):
    print("%d. Floorplan: %s"%(
        index + 1,
        floorplan.name,
    ))

    for index, track in enumerate(floorplan.tracks):
        print(" --> %d. Track: %s"%(
            index + 1,
            track.name
        ))

robot.get_state()


# Clean just a set of rooms
#floorplan = account.floorplans.pop()
#tracks = list(floorplan.tracks)[2:4]
#robot.start_cleaning(floorplan, tracks)

# Clean whole floor
#floorplan = account.floorplans.pop()
#robot.start_cleaning(floorplan)



import os
from dataclasses import dataclass
from typing import List, Union

@dataclass(init=False, frozen=True)
class Vendor:
    name: str
    friendly_name: str
    endpoint: str
    auth_endpoint: str
    mime_version: str = "application/vnd.neato.orbital-http.v1+json"

class Neato(Vendor):
    name = "neato"
    friendly_name: str = "Neato"
    endpoint = "https://orbital.neatocloud.com/"
    auth_endpoint = "https://orbital.neatocloud.com/vendors/neato/sessions"
    mime_version: str = "application/vnd.neato.orbital-http.v1+json"



from enum import Enum

class RobotAbilityEnum(str, Enum):
    STATE_SHOW = "state.show"
    INFO = "info.robot"
    FIND_ME = "utilities.find_me"
    CLEANING_PAUSE = "cleaning.pause"
    CLEANING_RESUME = "cleaning.resume"
    RETURN_TO_BASE = "navigation.return_to_base"

class RobotStateEnum(Enum):
    BUSY = 'busy'
    IDLE = 'idle'
    PAUSED = 'paused'

    @staticmethod
    def from_str(state: str):
        if state == 'busy':
            return RobotStateEnum.BUSY
        elif state == 'idle':
            return RobotStateEnum.IDLE
        elif state == 'paused':
            return RobotStateEnum.PAUSED
        else:
            raise NotImplementedError('RobotState %s'%state)


class RobotActionEnum(Enum):
    INVALID = 'invalid'
    CLEANING = 'cleaning'
    DOCKING = 'docking'

    @staticmethod
    def from_str(action: str):
        if action == 'invalid':
            return RobotActionEnum.INVALID
        elif action == 'cleaning':
            return RobotActionEnum.CLEANING
        elif action == 'docking':
            return RobotActionEnum.DOCKING
        else:
            raise NotImplementedError('RobotAction %s'%action)


class RobotBaseTypeEnum(Enum):
    STANDARD = 'standard'

    @staticmethod
    def from_str(type: str):
        if type == 'standard':
            return RobotBaseTypeEnum.STANDARD
        else:
            raise NotImplementedError('RobotBaseType %s'%type)


class RobotBagStatusEnum(Enum):
    BAG_OK = 'bag_ok'

    @staticmethod
    def from_str(status: str):
        if status == 'bag_ok':
            return RobotBagStatusEnum.BAG_OK
        else:
            raise NotImplementedError


class BaseTypeEnum(str, Enum):
    STANDARD = "standard"

    @staticmethod
    def from_str(type: str):
        if type == 'standard':
            return TrackTypeEnum.STANDARD
        else:
            raise NotImplementedError('BaseType %s'%type)


class TrackTypeEnum(str, Enum):
    CLEANING = "cleaning"
    NOGO = "no-go"

    @staticmethod
    def from_str(type: str):
        if type == 'cleaning':
            return TrackTypeEnum.CLEANING
        if type == 'no-go':
            return TrackTypeEnum.NOGO
        else:
            raise NotImplementedError('TrackType %s'%type)


class CleaningModeEnum(str, Enum):
    ECO = "eco"
    MAX = "max"
    TURBO = "turbo"

    @staticmethod
    def from_str(mode: str):
        if mode == 'eco':
            return CleaningModeEnum.ECO
        elif mode == 'max':
            return CleaningModeEnum.MAX
        elif mode == 'turbo':
            return CleaningModeEnum.TURBO
        else:
            raise NotImplementedError('CleaningMode %s'%mode)


class NavigationModeEnum(str, Enum):
    NORMAL = "normal"

    @staticmethod
    def from_str(mode: str):
        if mode == 'normal':
            return NavigationModeEnum.NORMAL
        else:
            raise NotImplementedError('NavigationMode %s'%mode)

class RobotErrorEnum(str, Enum):
    DUSTBIN_MISSING = 'dustbin_missing'



import binascii
import json
import os
import os.path
import requests
import logging
from typing import Callable, Dict, Optional

from .neato import Vendor, Neato
from .exception import MyNeatoException, MyNeatoLoginException, MyNeatoRobotException

try:
    from urllib.parse import urljoin
except ImportError:
    from urlparse import urljoin

_LOGGER = logging.getLogger(__name__)

class Session:
    def __init__(self, vendor: Vendor):
        """Initialize the session."""
        self.vendor = vendor
        self.endpoint = vendor.endpoint
        self.headers = {"Accept": vendor.mime_version}
        self.access_token = None
        self.is_active = False

    def get(self, path, **kwargs):
        """Send a GET request to the specified path."""
        raise NotImplementedError

    def urljoin(self, path):
        return urljoin(self.endpoint, path)

    def generate_headers(
        self, custom_headers: Optional[Dict[str, str]] = None
    ) -> Dict[str, str]:
        """Merge self.headers with custom headers if necessary."""
        if not custom_headers:
            return self.headers

        return {**self.headers, **custom_headers}


class OrbitalPasswordSession(Session):
    def __init__(self, email: str, password: str, access_token: str = None, vendor: Vendor = Neato()):
        super().__init__(vendor=vendor)
        self._email = email
        self._password = password
        self.access_token = access_token

        if self.access_token == None:
            self._login(email, password)

    def _login(self, email: str, password: str) -> None:
        """
        Login to your MyNeato account
        """
        _LOGGER.debug("Activating session")

        try:
            response = requests.post(
                urljoin(self.endpoint, "vendors/neato/sessions"),
                json={
                    "email": email,
                    "password": password,
                },
                headers=self.headers,
            )

            response.raise_for_status()

            self.access_token = response.json()["token"]

            self.is_active = True
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
            requests.exceptions.Timeout,
        ) as ex:
            if (
                isinstance(ex, requests.exceptions.HTTPError)
                and ex.response.status_code == 403
            ):
                raise MyNeatoLoginException(
                    "Unable to login to MyNeato account. check account credentials."
                ) from ex
            raise MyNeatoRobotException("Unable to connect to MyNeato API.") from ex

    def get(self, path, **kwargs):
        if not self.is_active:
            self._login(self._email, self._password)

        self.headers["Authorization"] = "Token %s" % self.access_token

        url = self.urljoin(path)
        headers = self.generate_headers(kwargs.pop("headers", None))
        try:
            response = requests.get(url, headers=headers, **kwargs)
            response.raise_for_status()
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
            requests.exceptions.Timeout,
        ) as ex:
            raise MyNeatoException("Unable to connect to MyNeato servers.") from ex

        return response



from typing import List
from enum import Enum
from .enum import RobotStateEnum, RobotActionEnum, RobotBaseTypeEnum, RobotBagStatusEnum

class RobotState:
    def __init__(self, action: RobotActionEnum, state: RobotStateEnum, available_commands: List[str] = []):
        self.action = action
        self.state = state
        self.available_commands = available_commands
        self.cleaning_center = None
        self.details = None


class RobotStateDetail:
    def __init__(
        self,
        base_type: RobotBaseTypeEnum,
        charge: int,
        is_charging: bool,
        is_docked: bool,
        is_quickboost: bool,
        quickboot_estimate: int = -1
    ):
        self._base_type = base_type
        self.charge = charge
        self.is_charging = is_charging
        self.is_docked = is_docked
        self.is_quickboost = is_quickboost
        self.quickboot_estimate = quickboot_estimate


class RobotStateCleaningCenter:
    def __init__(self, bag_status: RobotBagStatusEnum, base_error: None, is_extracting: bool):
        self._bag_status = bag_status
        self.base_error = base_error
        self.is_extracting = is_extracting



import re
import logging
import requests

from voluptuous import (
    ALLOW_EXTRA,
    All,
    Any,
    Extra,
    MultipleInvalid,
    Range,
    Required,
    Schema,
    Coerce,
    Equal
)
from enum import Enum

from .neato import Neato
from .enum import CleaningModeEnum, NavigationModeEnum, RobotAbilityEnum, RobotBaseTypeEnum, RobotBagStatusEnum, RobotActionEnum, RobotStateEnum
from .floorplan import Floorplan, Track
from .robot_state import RobotState, RobotStateDetail, RobotStateCleaningCenter
from .exception import MyNeatoRobotException

_LOGGER = logging.getLogger(__name__)

RUN_SCHEMA = Schema(
    {
        "map": {
            "nogo_enabled": bool,
            "rank_id": str,
            "track_id": Any(str, None)
        },
        "settings": {
            "mode": str,
            "navigation_mode": Coerce(CleaningModeEnum)
        }
    }
)

CLEANING_SCHEMA = Schema(
    {
        "ability": str,
        "force_floorplan": bool,
        "runs": [RUN_SCHEMA]
    }
)

ABILITY_SCHEMA = Schema(
    {
        'ability': str,
    }
)

STATE_SCHEMA = Schema(
    {
        Required('ability'): Equal(RobotAbilityEnum.STATE_SHOW),
        'action': str,
        'autonomy_states': {
            'active_cleaning_after_suspended': int,
            'active_cleaning_session': int,
            'cleaning_start': int,
            'docking': int,
            'docking_for_suspended': int,
            'docking_successful': int,
            'docking_successful_suspended': int,
            'docking_verify_base': int,
            'started_on_base': bool,
            'suspended_charging_start': int,
            'undocking': int,
            'undocking_after_suspended': int
        },
        'available_commands': {
            'cancel': bool,
            'pause': bool,
            'resume': bool,
            'return_to_base': bool,
            'start': bool
        },
        'cleaning_center': {
            'bag_status': str,
            'base_error': Any(str, None),
            'is_extracting': bool
        },
        'details': {
            'base_type': str,
            'charge': int,
            'is_charging': bool,
            'is_docked': bool,
            'is_quickboost': bool,
            'quickboost_estimate': int
        },
        'errors': Any(None),
        'state': str
    }
)

ROBOT_INFO_SCHEMA = Schema(
    {
        'ability': str,
        'firmware': str,
        'serial_number': str,
    }
)

class Robot:
    """Data and methods for interacting with a Neato vacuum robot"""

    def __init__(
        self,
        session,
        serial,
        id,
        user_id,
        name,
        endpoint,
        vendor_code,
        vendor=Neato,
    ):
        self._session = session
        self.name = name
        self.vendor = vendor
        self._vendor_code = vendor_code
        self.serial = serial
        self.id = id
        self.user_id = user_id
        self.endpoint = endpoint
        self.model_name = None
        self.firmware = None
        self.timezone = None
        self.birth_date = None

        self._url = "{endpoint}/vendors/{vendor_code}/robots/{serial}/messages".format(
            endpoint=re.sub(":\d+", "", endpoint.rstrip('/')),
            vendor_code=vendor_code,
            serial=self.serial,
        )
        self._headers = session.headers
        self._headers["Accept"] = vendor.mime_version

    def __str__(self):
        return "Name: %s, Serial: %s, ID: %s UserID: %s" % (
            self.name,
            self.serial,
            self.id,
            self.user_id,
        )

    def _message(self, message: str, json: dict, schema: Schema):
        """
        Sends message to robot with data from parameter 'json'
        :param json: dict containing data to send
        :return: server response
        """
        try:
            response = requests.post(
                self._url + "?ability=%s"%message,
                json=json,
                headers=self._headers,
            )
            response.raise_for_status()
            schema(response.json())
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.HTTPError,
        ) as ex:
            _LOGGER.warning("Unable to communicate with robot: %s"%(
                ex
            ))
            raise MyNeatoRobotException("Unable to communicate with robot") from ex
        except MultipleInvalid as ex:
            _LOGGER.warning(
                "Invalid response from %s: %s. Got: %s", self._url, ex, response.json()
            )

        return response

    def start_cleaning(
        self,
        floorplan: Floorplan,
        tracks: list[Track] = None,
        cleaning_mode = CleaningModeEnum.ECO,
        nogo_enabled = True
    ):
        ability_name = "cleaning.start"
        runs = []

        if tracks == None:
            runs.append({
                "map": {
                    "nogo_enabled": True,
                    "rank_id": floorplan.rank_uuid,
                    "track_id": None
                },
                "settings": {
                    "mode": cleaning_mode.value,
                    "navigation_mode": NavigationModeEnum.NORMAL.value
                }
            })
        else:
            for track in tracks:
                runs.append({
                    "map": {
                        "nogo_enabled": True,
                        "rank_id": floorplan.rank_uuid,
                        "track_id": track.uuid
                    },
                    "settings": {
                        "mode": track.cleaning_mode.value if track.cleaning_mode != None else None,
                        "navigation_mode": NavigationModeEnum.NORMAL.value
                    }
                })

        json = {
            "ability": ability_name,
            "force_floorplan": False,
            "runs": runs
        }

        response = self._message(ability_name, json, CLEANING_SCHEMA)
        result = response.json().get("ability", None)
        if result != ability_name:
            _LOGGER.warning(
                "Result of robot.%s is not ok: %s", ability_name, result
            )

        return {
            result: result == ability_name,
            response: response
        }

    def _base_message(self, message: str, schema: Schema):
        ability_name = message
        json = {
            "ability": message,
        }

        response = self._message(message, json, schema)
        result = response.json().get("ability", None)
        if result != message:
            _LOGGER.warning(
                "Result of robot.%s is not ok: %s, alert: %s", message, result
            )

        return {
            "success": result == ability_name,
            "response": response
        }

    def pause_cleaning(self) -> bool:
        result = self._base_message(RobotAbilityEnum.CLEANING_PAUSE.value, ABILITY_SCHEMA)

        return result["success"]

    def get_state(self) -> RobotState:
        result = self._base_message(RobotAbilityEnum.STATE_SHOW.value, STATE_SCHEMA)
        response = result["response"]

        json = response.json()

        state = RobotState(
            RobotActionEnum(json["action"]),
            RobotStateEnum(json["state"]),
        )

        state.cleaning_center = RobotStateCleaningCenter(
            RobotBagStatusEnum(json["cleaning_center"]["bag_status"]),
            json["cleaning_center"]["base_error"],
            json["cleaning_center"]["is_extracting"],
        )

        state.details = RobotStateDetail(
            RobotBaseTypeEnum(json["details"]["base_type"]),
            json["details"]["charge"],
            json["details"]["is_charging"],
            json["details"]["is_docked"],
            json["details"]["is_quickboost"],
            json["details"]["quickboost_estimate"],
        )

        for command, available in json["available_commands"].items():
            if available:
                state.available_commands.append(command)

        return state

    def info_robot(self):
        result = self._base_message(RobotAbilityEnum.INFO.value, ROBOT_INFO_SCHEMA)
        response = result["response"]

        return response.json()

    def resume_cleaning(self) -> bool:
        result = self._base_message(RobotAbilityEnum.CLEANING_RESUME.value, ABILITY_SCHEMA)

        return result["success"]

    def cancel_cleaning(self) -> bool:
        pause_result = self.pause_cleaning()
        return_to_base_result = self.return_to_base()

        return pause_result and return_to_base_result

    def return_to_base(self) -> bool:
        result = self._base_message(RobotAbilityEnum.RETURN_TO_BASE.value, ABILITY_SCHEMA)

        return result["success"]

    def find_me(self) -> bool:
        result = self._base_message(RobotAbilityEnum.FIND_ME.value, ABILITY_SCHEMA)

        return result["success"]

    @property
    def state(self):
        return self.get_state()



import logging
import os
import shutil
from typing import List

import requests

from .exception import MyNeatoRobotException, MyNeatoUnsupportedDevice
from .session import Session
from .robot import Robot
from .floorplan import Floorplan

from voluptuous import (
    ALLOW_EXTRA,
    All,
    Any,
    Extra,
    MultipleInvalid,
    Optional,
    Range,
    Required,
    Schema,
    Url,
)

_LOGGER = logging.getLogger(__name__)

USERDATA_SCHEMA = Schema(
    {
      Required("country_code"): str,
      Required("email"): str,
      Required("id"): str,
      Required("locale"): str,
      Required("first_name"): str,
      Required("last_name"): str,
      "newsletter": bool,
      "state_region": Any(str, None),
      "vendor": str,
      "verified_at": Any(str, None)
    },
    extra=ALLOW_EXTRA,
)

ROBOT_SCHEMA = Schema(
    {
        Required("id"): str,
        Required("user_id"): str,
        Required("serial"): str,
        Required("name"): str,
        "model_name": Any(str, None),
        "firmware": Any(str, None),
        "timezone": Any(str, None),
        "mac_address": Any(str, None),
        "birth_date": Any(str, None),
        "vendor": Any(str, None),
    },
    extra=ALLOW_EXTRA,
)

FLOORPLAN_SCHEMA = Schema(
    {
        Required("floorplan_uuid"): str,
        Required("rank_uuid"): str,
        Required("name"): Any(str, None),
        "map_versions_count": int,
    },
    extra=ALLOW_EXTRA,
)

class Account:
    def __init__(self, session: Session):
        """Initialize the account data."""
        self._session = session
        self._robots = []
        self._floorplans = []
        self._userdata = set()

    @property
    def robots(self) -> List[Robot]:
        """
        Return set of robots for logged in account.

        :return:
        """
        if not self._robots:
            self.refresh_robots()

        return self._robots

    @property
    def floorplans(self) -> List[Floorplan]:
        """
        Return set of floorplans for logged in account.
        """
        if not self._floorplans:
            self.refresh_floorplans()

        return self._floorplans

    @property
    def userdata(self):
        """
        Return the user
        """
        if not self._userdata:
            self.get_userdata()

        return self._userdata

    def refresh_robots(self):
        """
        Get information about robots connected to account.

        :return:
        """

        resp = self._session.get("users/me/robots")

        self._robots = []
        for robot in resp.json():
            _LOGGER.debug("Create Robot: %s", robot)
            try:
                ROBOT_SCHEMA(robot)
                robot_object = Robot(
                    session=self._session,
                    serial=robot["serial"],
                    id=robot["id"],
                    user_id=robot["user_id"],
                    name=robot["name"],
                    endpoint=self._session.endpoint,
                    vendor_code=robot["vendor"],
                    vendor=self._session.vendor,
                )
                robot_object.birth_date = robot['birth_date']
                robot_object.firmware = robot['firmware']
                robot_object.model_name = robot['model_name']
                robot_object.timezone = robot['timezone']

                self._robots.append(robot_object)
            except MultipleInvalid as ex:
                # Robot was not described accordingly by neato
                _LOGGER.warning(
                    "Bad response from robots endpoint: %s. Got: %s", ex, robot
                )
                continue
            except MyNeatoUnsupportedDevice:
                # Robot does not support home_cleaning service
                _LOGGER.warning("Your robot %s is unsupported.", robot["name"])
                continue
            except MyNeatoRobotException:
                # The state of the robot could not be received
                _LOGGER.warning("Your robot %s is offline.", robot["name"])
                continue

    def refresh_floorplans(self):
        _LOGGER.debug("Getting floorplans")

        for robot in self.robots:
            self.get_floorplan(robot)

    def get_floorplan(self, robot: Robot):
        _LOGGER.debug("Getting floorplan for %s", robot.name)

        resp = self._session.get("/robots/%s/floorplans"%robot.id)

        self._floorplans = []
        for floorplan in resp.json():
            FLOORPLAN_SCHEMA(floorplan)
            floorplan_object = Floorplan(
                session = self._session,
                uuid = floorplan["floorplan_uuid"],
                name = floorplan["name"],
                rank_uuid = floorplan["rank_uuid"],
            )
            self._floorplans.append(floorplan_object)

    def get_userdata(self):
        resp = self._session.get("/users/me")

        json = resp.json()
        USERDATA_SCHEMA(json)
        self._userdata = json



from .account import Account
from .floorplan import Floorplan
from .neato import Neato
from .robot import Robot, CleaningModeEnum, NavigationModeEnum
from .robot_state import RobotState, RobotStateDetail, RobotStateCleaningCenter
from .session import Session, OrbitalPasswordSession
from .enum import TrackTypeEnum, CleaningModeEnum, RobotStateEnum, RobotAbilityEnum, RobotActionEnum, RobotBaseTypeEnum, BaseTypeEnum, NavigationModeEnum
from .version import __version__
from .exception import MyNeatoLoginException, MyNeatoRobotException, MyNeatoException



import logging

from .session import Session
from .enum import TrackTypeEnum, CleaningModeEnum

from voluptuous import (
    ALLOW_EXTRA,
    All,
    Any,
    Extra,
    MultipleInvalid,
    Optional,
    Range,
    Required,
    Schema,
    Url,
    Coerce,
)

_LOGGER = logging.getLogger(__name__)

TRACK_SCHEMA = Schema(
    Any(
        {
            "track_uuid": str,
            "name": Any(str, None),
            "icon_id": Any(str, None),
            "type": Coerce(TrackTypeEnum),
            "binary": str,
            "cleaning_mode": Any(Coerce(CleaningModeEnum), None),
            "inserted_at": str,
            "updated_at": str,
        },
        extra=ALLOW_EXTRA,
    ),
    extra=ALLOW_EXTRA,
)

class Floorplan:
    def __init__(self, session: Session, uuid: str, name: str | None, rank_uuid: str):
        self._session = session
        self.name = name
        self.uuid = uuid
        self.rank_uuid = rank_uuid
        self._tracks = set()

    @property
    def tracks(self):
        """
        Return set of tracks for this floorplan

        :return:
        """
        if not self._tracks:
            self.refresh_tracks()

        return self._tracks

    def __str__(self):
        return "Name: %s, UUID: %s, RankID: %s" % (
            self.name,
            self.uuid,
            self.rank_uuid,
        )

    def refresh_tracks(self):
        resp = self._session.get("maps/floorplans/%s/tracks"%(self.uuid))

        for track in resp.json():
            if track["name"] == None:
                continue

            try:
                cleaning_mode = None
                if None != track["cleaning_mode"]:
                    cleaning_mode = CleaningModeEnum(track["cleaning_mode"])

                TRACK_SCHEMA(track)
                track_object = Track(
                    floorplan=self,
                    uuid=track["track_uuid"],
                    name=track["name"],
                    type=track["type"],
                    cleaning_mode=cleaning_mode
                )

                self._tracks.add(track_object)
            except MultipleInvalid as ex:
                _LOGGER.warning(
                    "Bad response from tracks endpoint: %s. Got: %s", ex, track
                )
                continue


class Track:
    def __init__(self, floorplan: Floorplan, uuid: str, name: str, type: str, cleaning_mode: CleaningModeEnum):
        """"""
        self.floorplan = Floorplan
        self.uuid = uuid
        self.name = name
        self.type = type
        self.cleaning_mode = cleaning_mode
